# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cache_data.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='cache_data.proto',
  package='ultron.cache_data',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\x10\x63\x61\x63he_data.proto\x12\x11ultron.cache_data\">\n\tCacheData\x12\x0f\n\x07session\x18\x01 \x02(\x0c\x12\x0b\n\x03key\x18\x02 \x02(\x0c\x12\x13\n\x0b\x64\x61ta_stream\x18\x03 \x02(\x0c')
)




_CACHEDATA = _descriptor.Descriptor(
  name='CacheData',
  full_name='ultron.cache_data.CacheData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='session', full_name='ultron.cache_data.CacheData.session', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='key', full_name='ultron.cache_data.CacheData.key', index=1,
      number=2, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data_stream', full_name='ultron.cache_data.CacheData.data_stream', index=2,
      number=3, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=39,
  serialized_end=101,
)

DESCRIPTOR.message_types_by_name['CacheData'] = _CACHEDATA
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CacheData = _reflection.GeneratedProtocolMessageType('CacheData', (_message.Message,), dict(
  DESCRIPTOR = _CACHEDATA,
  __module__ = 'cache_data_pb2'
  # @@protoc_insertion_point(class_scope:ultron.cache_data.CacheData)
  ))
_sym_db.RegisterMessage(CacheData)


# @@protoc_insertion_point(module_scope)
